#ifndef __COMPOSITEPATHALGEBRAS__ 
#define __COMPOSITEPATHALGEBRAS__ 

#include <iostream>
#include <sstream>
#include <string>
#include <limits>
//#include <list>
#include <set>
#include "Timer.h"
#include "PathAlgebras.h"
#include "SetAlgebras.h"
#include "VectorAlgebras.h"

using namespace std;

template<class Type1, class Type2>
class CompositePair {
public:
	Type1 first;
	Type2 second;
	CompositePair(const Type1 &f = Type1::zero, const Type2 &s = Type2::zero) : first (f), second (s) {}
	bool lex_LE(const CompositePair &s) const {
		return first.lex_LE(s.first);
	}
	string ToString () const {
		string s = "(" + first.ToString() + "," + second.ToString() + ")";
		return s;
	}
	string ToDelimitedString (const char &delim = ' ') const {
		return first.ToDelimitedString() + " " + second.ToDelimitedString();
	}
};


template<class Type1, class Type2>
ostream& operator<< (ostream &os, const CompositePair<Type1,Type2> &p) {
	os << p.ToString();
	return os;
}


template<class Type1, class Type2>
class CompositeSetPathAlgebra {
	typedef std::function<bool (const CompositePair<Type1,Type2>&, const CompositePair<Type1,Type2>&)> CompositePairComparator;
	typedef set<CompositePair<Type1,Type2>, CompositePairComparator> CompositePairSet;
	CompositePairSet element;
	CompositePair<Type1,Type2> setElementPrototype; 
public:
	const static CompositeSetPathAlgebra zero;
	const static CompositeSetPathAlgebra one;

	CompositeSetPathAlgebra() : element ( [](const CompositePair<Type1,Type2>& x, const CompositePair<Type1,Type2>& y) {return x.lex_LE(y);} ) {}

	CompositeSetPathAlgebra(const CompositePair<Type1,Type2> &e ) 
			: element ( [](const CompositePair<Type1,Type2>& x, const CompositePair<Type1,Type2>& y) {return x.lex_LE(y);} ) {
		element.clear();
//		Insert(e);
		setElementPrototype = e;
	}
	
	CompositeSetPathAlgebra(const Type1 &t1, const Type2 &t2 ) 
			: element ( [](const CompositePair<Type1,Type2>& x, const CompositePair<Type1,Type2>& y) {return x.lex_LE(y);} ) {
		element.clear();
		setElementPrototype = CompositePair(t1,t2);
	}
	
	const CompositeSetPathAlgebra& operator+ (const CompositeSetPathAlgebra &s) const {
		if (s.IsZero()) return (*this);
		if (this->IsZero()) return s;
		static CompositeSetPathAlgebra *r = new CompositeSetPathAlgebra();
		r->element.clear();
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			r->Insert(*i);
		}
		for (auto i = s.element.begin(); i !=s.element.end(); ++i ) {
			r->Insert(*i);
		}
		return *r;
	}

	CompositeSetPathAlgebra& operator+= (const CompositeSetPathAlgebra &s) {
		if (s.IsZero()) return (*this);
		if (this->IsZero()) {
			return (*this = s);
		}
		for (auto i = s.element.begin(); i !=s.element.end(); ++i ) {
			Insert(*i);
		}
		return *this;
	}

	const CompositeSetPathAlgebra& operator* (const CompositeSetPathAlgebra &s) const {
		if ( s.IsZero() || this->IsZero() ) return zero;
		if ( s.IsOne() ) return (*this);
		if ( this->IsOne() ) return s;
		static CompositeSetPathAlgebra *r = new CompositeSetPathAlgebra();
		r->element.clear();
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			for (auto j = s.element.begin(); j !=s.element.end(); ++j ) {
				r->Insert(CompositePair(i->first * j->first, i->second * j->second));
			}
		}
		return *r;
	}

	bool operator==(const CompositeSetPathAlgebra &s) const {
		if (element.size() != s.element.size()) return false;
		for (auto i = element.begin(); i !=element.end(); ++i ) {
			bool r = false;
			for (auto j = s.element.begin(); j !=s.element.end(); ++j ) {
				if (i->first == j->first) {
					r = true;
					break;
				}
			}
			if (r == false) return false;
		}
		return true;
	}

	bool operator<=(const CompositeSetPathAlgebra &s) const {
		return   ( ( (*this) + s) == s ); 
	}

	bool lex_LE(const CompositeSetPathAlgebra &s) const {
		auto i = element.begin();
		auto j = s.element.begin();
		while (i != element.end() && j != s.element.end()) {
			if (i->first.lex_LE(j->first)) return true;
			else if (j->first.lex_LE(i->first)) return false;
			++i; ++j;
		}
		return false; 
	}

	string ToString() const {
		string s = "{";
		for(auto i = element.begin(); i != element.end(); ++i) {
			s += "|" + i->first.ToString() + "*" + i->second.ToString() + "|,";
		}
		if (s[s.size()-1] == ',') s[s.size()-1] = '}';
		else s += "}";
		return s;
	}

	string ToDelimitedString(char delim = ' ') const {
		string s = "";
		for(auto i = element.begin(); i != element.end(); ++i) {
			s += i->first.ToDelimitedString() + delim + i->second.ToDelimitedString() + "\t";
		}
		return s;
	}

	virtual void FromString(string & s) {
		cerr << "CompositeSetAlgebra::FromString not implemented yet." << endl; cerr.flush();
		exit(1);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		element.clear();
		EraseLeadingBlanks(s);
		setElementPrototype.first.FromDelimitedString(s); 
		setElementPrototype.second.FromDelimitedString(s); 
		Insert(setElementPrototype);
	}

	void Insert(const CompositePair<Type1,Type2> &e)  {
		if (element.size() == 0) {
			element.insert(e);
			return;
		}
		for (auto j = element.begin(); j !=element.end(); ++j) {
			if (e.first == j->first) { 
				CompositePair t(e.first,j->second + e.second);
				element.erase(j);
				element.insert(t);
				return;
			}
			else if (e.first <= j->first) { //(e.first < j->first)
				return;
			}
			else if (j->first <= e.first) { //(e.first > j->first)
				while (j !=element.end()) {
					if (j->first <= e.first) { //(e.first > j->first) 
						j = element.erase(j);
					}
					else  ++j;
				}
				element.insert(e);
				return;
			}
		}
		element.insert(e);
		return;
	}

	void Erase(const CompositePair<Type1,Type2> &e)  {
		element.erase(e);
	}

	const CompositePairSet& GetElement() const {
		return element;
	}

	const CompositeSetPathAlgebra & Zero() const {
		return zero;
	}
	const CompositeSetPathAlgebra & One() const {
		return one;
	}
	bool IsZero() const {
		return (element.size() == 0);
	}
	bool IsOne() const {
		if (element.size() != 1) return false;
		auto i = element.begin();
		return ( (i->second == Type2::one) && (i->first == Type1::one) ); 
	}
};

template<class Type1, class Type2>
const CompositeSetPathAlgebra<Type1,Type2> CompositeSetPathAlgebra<Type1,Type2>::zero ; 

template<class Type1, class Type2>
const CompositeSetPathAlgebra<Type1,Type2> CompositeSetPathAlgebra<Type1,Type2>::one (Type1::one,Type2::one) ;

template<class Type1, class Type2>
ostream& operator<< (ostream &os, const CompositeSetPathAlgebra<Type1,Type2> &s) {
	os << s.ToString();
	return os;
}

template<class Type1, class Type2>
string ToString (const pair<Type1,Type2> &p) {
	string s = "(" + p.first.ToString() + "," + p.second.ToString() + ")";
	return s;
}

template<class Type1, class Type2>
string ToDelimitedString (const pair<Type1,Type2> &p, const char &delim = ' ') {
	return p.first.ToDelimitedString() + " " + p.second.ToDelimitedString();
}

template<class Type1, class Type2>
ostream& operator<< (ostream &os, const pair<Type1,Type2> &p) {
	os << ToString(p);
	return os;
}




template<class Type1, class Type2>
PathAlgebraBaseType<CompositeSetPathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > > 
	CompositeAlgebra(const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2) {
	return Algebra(CompositeSetPathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > (e1,e2) );
}

template<class Type1, class Type2>
PathAlgebraBaseType<CompositeSetPathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > > 
	C(const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2) {
	return Algebra(CompositeSetPathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > 
					(CompositePair(e1,e2)) );
}


template<class Type1, class Type2>
auto MaxMinAlgebra(const PathAlgebraBaseType<CompositeSetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1> >,PathAlgebraBaseType<Type2> > > &e) {
 	auto sa = e.GetElement(); //CompositeSetAlgebra
  	auto lav = sa.GetElement(); // list of PathAlgebras
	if (lav.begin() == lav.end() ) {
		PrintError("MaxMin error: list empty.");
		return CompositePair<PathAlgebraBaseType<VectorNPathAlgebra<Type1> >, PathAlgebraBaseType<Type2> >();
	}
	auto i = lav.begin(); // iterator to PathAlgebra
	auto maxi = i;
	Type1 maxmin = Min(i->first);
	++i;
	while (i != lav.end()) {
		auto nextmax = Min(i->first);
		if (!(nextmax <= maxmin)) {
			maxmin = nextmax;
			maxi = i;
		}
		++i;
	}
	return (*maxi);
}


template<class Type1, class Type2>
auto MaxMinValue(const PathAlgebraBaseType<CompositeSetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1> >,PathAlgebraBaseType<Type2> > > &e) {
	return Min( MaxMinAlgebra(e).first );
}


#endif

