#ifndef __MATRIXPATHALGEBRAS__
#define __MATRIXPATHALGEBRAS__

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <limits>
#include <chrono>
#include "PathAlgebras.h"
#include "Timer.h"

using namespace std;

template<class Type>
class Matrix {
	vector<vector<Type> > m;
	vector<Timer> vt;
public:
	Matrix<Type> (const Type& t, int n = 1) {
		m.reserve(n);
		m.resize(n);
		for (int i=0; i<n; ++i) {
			m[i].reserve(n);
			m[i].resize(n);
		}
		for (int i=0; i<n; ++i) {
			for (int j=0; j<n; ++j) {
				m[i][j] = t;
			}
		}
	}
	vector<vector<Type> >& GetElement() {
		return m;
	}

	void AddTimeLog(const Timer &t) {
		vt.push_back(t);
	}

	string ToString() const {
		ostringstream os;
		for (int i=0; i<m.size(); ++i) {
			for (int j=0; j<m[i].size(); ++j) {
				os << m[i][j] << "\t";
			}
			os << endl;
		}
		return os.str();
	}
	string ToDelimitedString(char delim = ' ') const {
		ostringstream os;
		for (int i=0; i<m.size(); ++i) {
			for (int j=0; j<m[i].size(); ++j) {
				os << m[i][j].ToDelimitedString() << endl;
			}
//			os << endl;
		}
		return os.str();
	}

	void FromFile (const string &filename) {
		string line;
		int i(0), j(0);
		ifstream f (filename);
		if (f.is_open()) {
			while (getline (f, line)) { 
				auto it =  find_if( line.begin() , line.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
				line.erase( line.begin() , it);
				if (line == "") continue;
				EraseComments(line);
				if (line == "") continue; 
				istringstream is(line);
				is >> i; 
				EraseReadCharacters(is);
				--i; 
				is >> j; 
				EraseReadCharacters(is);
				--j;
				if (i<0 || j<0 || i>=dim() || j>=dim()) {
					cerr << "Matrix::FromFile error: i or j out of range. " << endl; 
					exit(1);
				}
				line = is.str();
				m[i][j].FromDelimitedString(line);
			}
			f.close();
		}
		else {
			cerr << "Matrix::FromFile error: Unable to open file: " << filename << endl; 
			exit(1);
		}
	}

	void ToFile (const string &filefolder) {
		ofstream f (filefolder);
		if (f.is_open()) {
			ProblemParameters problem(filefolder);
			if (problem.solutionFileHeader != "") f << problem.solutionFileHeader << endl;

			for (int i = 0; i < vt.size(); ++i) { 
				f << vt[i] << endl;
			}

			f << "//Average Number of Vectors in Weak Closure: " << AvgVectorNumber(*this) << endl;
			f << endl;

			f << "//Matrix Closure section:" << endl;
			for (int i = 0; i < dim(); ++i) {
				for (int j = 0; j < dim(); ++j) {
					f << "//Closure element (" << (i+1) << "," << (j+1) << ")" << endl;
					if (!m[i][j].IsZero()) {
						f << (i+1) << " " << (j+1) << "\t" << m[i][j].ToDelimitedString() << endl;
					}
				}
				f << endl;
			}

			f << "//Matrix MaxMinValues section:" << endl;
			for (int i = 0; i < dim(); ++i) {
				for (int j = 0; j < dim(); ++j) {
					if (!m[i][j].IsZero()) {
						f << (i+1) << " " << (j+1) << "\t" << MaxMinValue(m[i][j]).ToDelimitedString() << endl;
					}
				}
				f << endl;
			}

			f << "//Matrix MaxMinAlgebras section:" << endl;
			for (int i = 0; i < dim(); ++i) {
				for (int j = 0; j < dim(); ++j) {
					if (!m[i][j].IsZero()) {
						f << (i+1) << " " << (j+1) << "\t" << MaxMinAlgebra(m[i][j]).ToDelimitedString() << endl;
					}
				}
				f << endl;
			}

			f.close();
		}
		else {
			cerr << "Matrix::ToFile error: Unable to open file: " << filefolder << endl; 
			exit(1);
		}
	}

	vector<Type>& operator[] (int i) {
		return m[i];
	}
	const vector<Type>& operator[] (int i) const {
		return m[i];
	}
	int dim () const {
		return m.size();
	}
	const Matrix<Type> & operator*(const Matrix<Type> &s) const {
		static Matrix<Type> *pr = new Matrix<Type> (s);
		Matrix<Type> &r = *pr;
		Type t;
		int n = dim();
		for(int i = 0; i < n; ++i) {
			for(int j = 0; j < n; ++j) {
				t = t.zero;
				for(int k = 0; k < n; ++k)
					t = t + m[i][k] * s[k][j];
				r[i][j] = t;
			}
		}
		return r;
	}
	bool operator== (const Matrix<Type> &s) const {
		int n = dim();
		for(int i = 0; i < n; ++i) {
			for(int j = 0; j < n; ++j) {
				if ( !(m[i][j] == s[i][j]) ) {
					cerr << "Matrices different at " << i << ", " << j << endl;
					cerr << m[i][j]  << endl;
					cerr << s.m[i][j]  << endl;
					return false;
				}
			}
		}
		return true;
	}
};

template <class Type>
Matrix<Type> ClosureStrong(const Matrix<Type> &A) { //A*
	return Astar(A);
}

template <class Type>
Matrix<Type> ClosureWeak(const Matrix<Type> &A) { //A+
	return ClosureStrong(A) * A;
}

template <class Type>
Matrix<Type> AtimesIplusAtoN(const Matrix<Type> &A) {
	return IplusAtoN(A) * A;
}

template <class Type>
Matrix<Type> IplusAtoN(const Matrix<Type> &A) {
	auto startTime = chrono::high_resolution_clock::now();
	int i, j, k;
	int n = A.dim();
	bool ok = true;
	Matrix<Type> IA = A;
	Matrix<Type> IAn (Type::zero, n);
	Type t;
	for(i = 0; i < n; ++i)	IA[i][i] = IA[i][i] + Type::one; // IA = I + A
	while (ok) {
		ok = false;
		for(i = 0; i < n; ++i) {
			for(j = 0; j < n; ++j) {
				IAn[i][j] = Type::zero;
				for(k = 0; k < n; ++k)	
					IAn[i][j] += IA[i][k] * IA[k][j];
				if( !ok && !(IAn[i][j] == IA[i][j]) ) {
					ok = true;
				}
			}
		}
		IA = IAn;
	}
	auto endTime = chrono::high_resolution_clock::now();
	auto durationTime = chrono::duration_cast<chrono::milliseconds>( endTime - startTime ).count();
	cerr << "IplusAtoN in " << durationTime << " ms." << endl;
	return IAn;
}

template <class Type>
Matrix<Type> AtimesIplusAtoNquick(const Matrix<Type> &A) {
	return IplusAtoNquick(A) * A;
}

template <class Type>
Matrix<Type> IplusAtoNquick(const Matrix<Type> &A) {
	auto startTime = chrono::high_resolution_clock::now();
	int i, j, k;
	int n = A.dim();
	bool ok = true;
	Matrix<Type> IA = A;
	Type t;
	for(i = 0; i < n; ++i)	IA[i][i] = IA[i][i] + Type::one; 
	while (ok) {
		ok = false;
		for(i = 0; i < n; ++i) {
			for(j = 0; j < n; ++j) {
				t = IA[i][j];
				for(k = 0; k < n; ++k)	
					t += IA[i][k] * IA[k][j];
				if( !(t == IA[i][j]) ) {
					ok = true;
					IA[i][j] = t;
				}
			}
		}
	}
	auto endTime = chrono::high_resolution_clock::now();
	auto durationTime = chrono::duration_cast<chrono::milliseconds>( endTime - startTime ).count();
	cerr << "IplusAtoNquick in " << durationTime << " ms." << endl;
	return IA;
}

template <class Type>
Matrix<Type> GaussSeidel(const Matrix<Type> &A) {
	Timer t("//GaussSeidel Execution Time: "); 
	int i, j, k;
	int n = A.dim();
	bool ok = true;
	Matrix<Type> B = A;
	Matrix<Type> Y = A;

	for(i = 0; i < n; ++i) 
		for(j = 0; j < n; ++j) 
			Y[i][j] = A[i][j] + (i == j ? Type::one : Type::zero);

	Type temp;
	while (ok) {
		ok = false;
		for(i = 0; i < n; ++i) {
			for(j = 0; j < n; ++j) {
				temp = B[i][j];
				for(k = 0; k < n; ++k)	temp = temp + A[i][k] * Y[k][j];
				if( !(temp == Y[i][j]) ) {
					Y[i][j] = temp;
					ok = true;
				}
			}
		}
	}
	t.end(); 
	Y.AddTimeLog(t); 
	return Y;
}

template <class Type>
Matrix<Type> Aplus(const Matrix<Type> &A) {
	auto startTime = chrono::high_resolution_clock::now();
	int i, j, k;
	int n = A.dim();
	bool ok = true;
	static Matrix<Type> C = Matrix<Type> (Type::zero,A.dim());
	static Matrix<Type> An = Matrix<Type> (Type::zero,A.dim());
	static Matrix<Type> Aprev = Matrix<Type> (Type::zero,A.dim());
	static Matrix<Type> S = Matrix<Type> (Type::zero,A.dim());
	for(i = 0; i < n; ++i)
		for(j = 0; j < n; ++j) {
			S[i][j] = A[i][j]; // S = A
			Aprev[i][j] = S[i][j]; // C = A
			An[i][j] = S[i][j]; // C = A
		}
	Type temp;
	while (ok) {
		ok = false;
		for(i = 0; i < n; ++i) {
			for(j = 0; j < n; ++j) {
				An[i][j] = Type::zero;
				for(k = 0; k < n; ++k)	{
					An[i][j] = An[i][j] + Aprev[i][k] * A[k][j];
				}
				temp = S[i][j] + An[i][j]; 
				if( !ok && !(temp == S[i][j]) ) { 
					ok = true;
				}
				S[i][j] = temp;
			}
		}
		Aprev = An;
	}
	auto endTime = chrono::high_resolution_clock::now();
	auto durationTime = chrono::duration_cast<chrono::milliseconds>( endTime - startTime ).count();
	cerr << "Aplus in " << durationTime << " ms." << endl;
   	return S;
}

template <class Type>
Matrix<Type> Astar(const Matrix<Type> &A) {
	auto startTime = chrono::high_resolution_clock::now();
	int i, j, k;
	int n = A.dim();
	bool ok = true;
	static Matrix<Type> An = Matrix<Type> (Type::zero,A.dim());
	static Matrix<Type> Aprev = Matrix<Type> (Type::zero,A.dim());
	static Matrix<Type> S = Matrix<Type> (Type::zero,A.dim());
	for(i = 0; i < n; ++i)
		for(j = 0; j < n; ++j) {
			S[i][j] = (i == j ? Type::one : Type::zero ); // S = I
			Aprev[i][j] = S[i][j]; // C = I
			An[i][j] = S[i][j]; // C = I
		}
	Type temp;
	while (ok) {
		ok = false;
		for(i = 0; i < n; ++i) {
			for(j = 0; j < n; ++j) {
				An[i][j] = Type::zero;
				for(k = 0; k < n; ++k)	{
					An[i][j] = An[i][j] + Aprev[i][k] * A[k][j]; 
				}
				temp = S[i][j] + An[i][j]; 
				if( !ok && !(temp == S[i][j]) ) { 
					ok = true;
				}
				S[i][j] = temp;
			}
		}
		Aprev = An;
	}
	auto endTime = chrono::high_resolution_clock::now();
	auto durationTime = chrono::duration_cast<chrono::milliseconds>( endTime - startTime ).count();
	cerr << "Astar in " << durationTime << " ms." << endl;
   	return S;
}


template<class Type>
ostream& operator<< (ostream &os, const Matrix<Type> &s) {
	for (int i=0; i<s.dim(); ++i) {
		for (int j=0; j<s[i].size(); ++j) {
			os << s[i][j] << "\t";
		}
		os << endl;
	}
	return os;
}

template<class Type>
Matrix<Type> MatrixAlgebra(const Type &t, int n = 1) {
	return Matrix<Type>(t, n);
}

template<class Type1, class Type2>
double AvgVectorNumber(const Matrix<PathAlgebraBaseType<CompositeSetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1> >,PathAlgebraBaseType<Type2> > > > &M) {
	double avg = 0.0;
	for (int i = 0; i < M.dim(); ++i) {
		for (int j = 0; j < M.dim(); ++j) {
			avg += M[i][j].GetElement().GetElement().size(); 
		}
	}
	return avg / (M.dim() * M.dim() - M.dim());
}

template<class Type1>
double AvgVectorNumber(const Matrix<PathAlgebraBaseType<SetPathAlgebra<PathAlgebraBaseType<VectorNPathAlgebra<Type1> > > > > &M) {
	double avg = 0.0;
	for (int i = 0; i < M.dim(); ++i) {
		for (int j = 0; j < M.dim(); ++j) {
			avg += M[i][j].GetElement().GetElement().size(); 
		}
	}
	return avg / (M.dim() * M.dim() - M.dim());
}


#endif

