#ifndef __PATHALGEBRAS__ 
#define __PATHALGEBRAS__ 

#include <iostream>
#include <sstream>
#include <string>
#include <limits>
#include "VectorAlgebras.h"

using namespace std;


template<class Type>
class PathAlgebraBaseType {
	Type	element;
public:
	static const PathAlgebraBaseType<Type> zero;
	static const PathAlgebraBaseType<Type> one;
	PathAlgebraBaseType<Type>(const Type &s = Type::zero) : element(s) {};

	const PathAlgebraBaseType<Type> & GetMyTypeReference (const PathAlgebraBaseType<Type> &s) {
		return dynamic_cast<const PathAlgebraBaseType<Type>&>(s);
	}
	
	const PathAlgebraBaseType<Type>& operator+ (const PathAlgebraBaseType<Type> &s) const {
		static PathAlgebraBaseType<Type> *r = new PathAlgebraBaseType<Type>(); 
		r->element = element + s.element;
		return *r; 
	}
	
	PathAlgebraBaseType<Type>& operator+= (const PathAlgebraBaseType<Type> &s) {
		element += s.element;
		return *this; 
	}
	
	const PathAlgebraBaseType<Type>& operator* (const PathAlgebraBaseType<Type> &s) const {
		static PathAlgebraBaseType<Type> *r = new PathAlgebraBaseType<Type>(); 
		r->element = element * s.element;
		return *r; 
	}

	bool operator==(const PathAlgebraBaseType<Type> &s)	const {
		return   (element == s.element) ; 
	}
	
	bool operator<=(const PathAlgebraBaseType<Type> &s)	const {
		return   (element <= s.element); 
	}

	bool lex_LE(const PathAlgebraBaseType<Type> &s) const {
		return   (element.lex_LE(s.element)); 
	}

	bool operator<(const PathAlgebraBaseType<Type> &s)	const {
		return   ( ((*this) <= s) && !((*this) == s) ); 
	}

	bool IsZero() const	{
		return (element == element.Zero()); 
	}
	bool IsOne() const	{
		return (element == element.One()); 
	}

	const PathAlgebraBaseType<Type>& Zero() const	{
		return zero;
	}
	const PathAlgebraBaseType<Type>& One() const	{
		return one;
	}

	const Type& GetElement() const	{
		return element; 
	}
	Type& Get() 	{
		return element; 
	}
	
	void SetElement(const Type &s) {
		element = s; 
	}

	string ToString() const {
		return "[" + element.ToString() + "]";
	}
	string ToDelimitedString(char delim = ' ') const {
		return element.ToDelimitedString();
	}
	void FromString(string & s) {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		istringstream is (s);
		ostringstream os;
		char c;
		is >> c;
		if (c != '[') exit(1);
		s = s.substr(1); is.str(s);
		element.FromString(s); is.str(s);
		it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		is >> c;
		if (c != ']') exit(1);
		s = s.substr(1); is.str(s);
	}

	void FromDelimitedString(string &s, char delim = ' ') {
		auto it =  find_if( s.begin() , s.end() , [](char ch){ return !isspace<char>(ch , locale::classic() ) ; } );
  		s.erase( s.begin() , it);
		if (s == "") return;
		element.FromDelimitedString(s);
	}

	void FromString(const char *pc) {
		string s (pc);
		return FromString(s);
	}

	void FromDelimitedString(const char *pc) {
		string s (pc);
		return FromDelimitedString(s);
	}

};

template<class Type>
const PathAlgebraBaseType<Type> PathAlgebraBaseType<Type>::zero(Type::zero) ;

template<class Type>
const PathAlgebraBaseType<Type> PathAlgebraBaseType<Type>::one (Type::one) ;

template<class Type>
ostream& operator<< (ostream &os, const PathAlgebraBaseType<Type> &s) {
	os << s.ToString();
	return os;
}

// double type needed for numeric_limits<double>::infinity. int does not have numeric_limits<int>::infinity.
typedef PathAlgebraBaseType<PSType<double>> P2;
typedef PathAlgebraBaseType<PSType<double>> PS;

typedef PathAlgebraBaseType<PLType<double>> P3;
typedef PathAlgebraBaseType<PLType<double>> PL;

typedef PathAlgebraBaseType<PRType> P4;
typedef PathAlgebraBaseType<PRType> PR;

typedef PathAlgebraBaseType<PCType<double>> P5;
typedef PathAlgebraBaseType<PCType<double>> PC;

typedef PathAlgebraBaseType<PAType<int>> P6;
typedef PathAlgebraBaseType<PAType<int>> PA;

typedef PathAlgebraBaseType<PEType<int>> P8;
typedef PathAlgebraBaseType<PEType<int>> PE;

template<class Type>
inline PathAlgebraBaseType<Type> Algebra(const Type &t) {
	return PathAlgebraBaseType<Type>(t);
}

template<class Type>
inline auto A(const Type &t) {
	return Algebra(t);
}


// VECTOR PART
template<class Type1, class Type2>
auto V(const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2) {
		return Algebra(VectorPathAlgebra<PathAlgebraBaseType<Type1>,PathAlgebraBaseType<Type2> > (e1,e2) );
	}

template<class Type1, class Type2, class Type3, class... Targs >
auto V(	const PathAlgebraBaseType<Type1> &e1, const PathAlgebraBaseType<Type2> &e2, 
				const PathAlgebraBaseType<Type3> &e3, Targs... Vargs) 
	{
		auto e12 = V (e1,e2);
		return V(e12,e3,Vargs...);
	}

template<class Type>
auto V(const int n, const PathAlgebraBaseType<Type> &e) {
		if (n <= 0) {
			cerr << "V error: n <= 0" << endl;
			exit(1);
		}
		return Algebra( VectorNPathAlgebra<PathAlgebraBaseType<Type>> (n,e) );
	}

template<class Type>
Type Max (const PathAlgebraBaseType<VectorNPathAlgebra<Type> > & e ) {
	return e.GetElement().Max();
}

template<class Type>
Type Min (const PathAlgebraBaseType<VectorNPathAlgebra<Type> > & e ) {
	return e.GetElement().Min();
}

// VECTOR PART END

#endif

